@@ -19,7 +19,7 @@ def link_tree( distribution_dir, target_dir='/', symbolic=True ):
             log.info( 'Creating directory: %s', dir_final )
             try:
                 os.makedirs( dir_final )
-            except Exception, err:
+            except Exception as err:
                 err.args += (dir_final,)
                 raise 
         for filename in filenames:
@@ -33,23 +33,30 @@ def link_tree( distribution_dir, target_dir='/', symbolic=True ):
                 ):
                     log.info( 'Up to date: %s', file_final )
                     continue
-                try:
-                    log.info( 'Replacing: %s', file_final )
-                    temp = file_final + '~'
-                    if os.path.exists( temp ):
-                        try:
-                            os.remove( temp )
-                        except Exception, err:
-                            err.args += (temp, )
-                            raise 
-                    link( file_source, temp )
-                    os.rename( temp, file_final )
-                except Exception, err:
-                    log.error( "Error linking %s: %s", file_final, err )
-                    raise err
+                log.info( 'Replacing: %s', file_final )
+                create_link( file_source, file_final, link )
             else:
                 log.info( 'Linking: %s', file_final )
-                link( file_source, file_final )
+                create_link( file_source, file_final, link )
+
+def create_link( file_source, file_final, link ):
+    try:
+        temp = file_final + '~'
+        if os.path.exists( temp ) or os.path.lexists( temp ):
+            try:
+                os.remove( temp )
+            except Exception as err:
+                err.args += (temp, )
+                raise 
+        try:
+            link( file_source, temp )
+        except OSError as err:
+            if link is os.link:
+                os.symlink( file_source, temp )
+        os.rename( temp, file_final )
+    except Exception as err:
+        log.error( "Error linking %s: %s", file_final, err )
+        raise err
     
 def get_options():
     """Creates the OptionParser used in :func:`main` """
@@ -66,11 +73,28 @@ def get_options():
         default = None,
         help="Directory in which to create the links",
     )
+    parser.add_option(
+        '-v','--verbose',
+        dest = 'verbose',
+        action="store_true",
+        default = False,
+        help="Perform verbose logging (INFO level), default: False",
+    )
+    parser.add_option(
+        '-H','--hard',
+        dest = 'hardlink',
+        action="store_true",
+        default = False,
+        help="Use hard-links instead of symlinks, default: False",
+    )
     return parser
 
 def main():
-    logging.basicConfig( level=logging.INFO )
     options,args = get_options().parse_args()
+    if options.verbose:
+        logging.basicConfig( level=logging.INFO )
+    else:
+        logging.basicConfig( level=logging.WARNING )
     if args:
         if not options.directory:
             options.directory = args[0]
@@ -78,5 +102,5 @@ def main():
         if args and not options.target:
             options.target = args[0]
             args = args[1:]
-    link_tree( options.directory, options.target )
+    link_tree( options.directory, options.target, symbolic = (not options.hardlink) )
     return 0
