@@ -3,13 +3,23 @@
 pipe = open( somefile, 'rb' ) | nbio.Process( ['grep','blue'] ) | nbio.Process( ['wc','-l'])
 
 """
-import os, select, fcntl, itertools, time, sys, popen2, signal, logging, subprocess, errno
+import os, select, fcntl, time, sys, logging, subprocess, errno, collections
 log = logging.getLogger( __name__ )
+from ._shims import bytes, unicode
+
+DEFAULT_READ_BLOCK = 1024*64
+
+def set_non_blocking( fh ):
+    """Set this file handle to non-blocking mode"""
+    current = fcntl.fcntl( fh, fcntl.F_GETFL )
+    fcntl.fcntl( fh, fcntl.F_SETFL, current | os.O_NONBLOCK )
+    return fh
 
 class _DidNothing( object ):
     """Returned when a subprocess didn't do anything"""
-    def __nonzero__( self ):
+    def __bool__( self ):
         return False
+    __nonzero__ = __bool__
 DID_NOTHING = _DidNothing()
 
 class NBIOError( RuntimeError ):
@@ -17,6 +27,16 @@ class NBIOError( RuntimeError ):
 class ProcessError( NBIOError ):
     """Called process returned an error code
     
+    Attributes:
+    
+        process -- the Process (if applicable) which raised the error
+        exitcode -- exit code from the process if available
+    """
+    process = None
+    exitcode = None
+class Timeout( NBIOError ):
+    """Called process did not complete within specified timeout
+    
     Attributes:
     
         process -- the Process (if applicable) which raised the error
@@ -35,7 +55,7 @@ class Pipe( object ):
     def __init__( self, *processes ):
         self.processes = []
         if processes:
-            first = self.append( self.get_component(processes[0]) )
+            self.append( self.get_component(processes[0]) )
             for process in processes[1:]:
                 self.__or__( process )
     def __repr__( self ):
@@ -49,7 +69,7 @@ class Pipe( object ):
         """Return the number of items in this pipe"""
         return len( self.processes )
         
-    def __iter__( self ):
+    def __iter__( self, timeout=None ):
         """Iterate over the processes in the pipe
         
         If the stdout/stderr of the processes is not captured, then we will 
@@ -60,6 +80,8 @@ class Pipe( object ):
         pipe will do a pause() for self.pause_on_silence (normally passed into the 
         __call__) before the next iteration.
         """
+        if timeout:
+            until = time.time() + timeout
         try:
             self.started = True
             iterables = [ iter(process) for process in self.processes ]
@@ -71,36 +93,46 @@ class Pipe( object ):
                     try:
                         child_result = iterable.next()
                         if child_result is DID_NOTHING:
+                            # we did nothing during this iteration
                             pass
                         elif child_result:
                             something_happened += 1
                             yield child_result
-                    except StopIteration, err:
+                    except StopIteration as err:
                         exhausted.append( iterable )
-                    except Exception, err:
+                    except Exception as err:
                         # Have the *first* item that failed raise the error, 
                         # and only if they don't, raise this error...
+                        log.error( 'Catching error: %s', err )
                         for item in self.processes:
                             item.check_exit()
                         raise
+                if not something_happened:
+                    for item in self.processes:
+                        if hasattr( item, 'poll' ):
+                            item.poll()
+                # check for a process/pipe/file failure...
                 for done in exhausted:
                     while done in iterables:
                         iterables.remove( done )
                 if not something_happened:
                     pause( self.pause_on_silence )
+                if timeout:
+                    if time.time() > until:
+                        raise Timeout( self, timeout )
         finally:
             self.kill()
-    def __call__( self, pause_on_silence=0.01 ):
+    def __call__( self, pause_on_silence=0.01, timeout=None ):
         """Iterate over this pipeline, returning combined results as a string
         """
         result = []
         self.pause_on_silence = pause_on_silence
         try:
-            for item in itertools.ifilter( bool, self ):
+            for item in self.__iter__( timeout=timeout ):
                 if item:
                     result.append( item )
             return "".join( result )
-        except Exception, err:
+        except Exception as err:
             err.output = result
             raise
     def append( self, process ):
@@ -127,19 +159,22 @@ class Pipe( object ):
     def __or__( self, other ):
         """Pipe our output into a process, callable or list"""
         other = self.get_component( other )
-        log.debug( 
-            'Hooking output of %s to input of %s', self.last, other 
-        )
-        self.last.iterables.append(
-            other.iter_write( self.last.iter_read() )
-        )
+        last = self.last 
+        # special cases for piping process-to-process 
+        both_processes = isinstance(last,Process) and isinstance( other, Process )
+        both_processes = False
+        if both_processes and not other.started and not last.started:
+            log.info( 'Direct pipe %s -> %s', last, other )
+            last.blocking_stdout()
+            other.direct_pipe( last.stdout, 'stdin' )
+        else:
+            last.iterables.append(
+                other.iter_write( self.last.iter_read() )
+            )
         self.append( other )
         return self
     def __ror__( self, other ):
         """Pipe output of other into our first item"""
-        log.debug(
-            'Hooking output of %s to input of %s', other, self.first 
-        )
         other = self.get_component( other )
         self.first.iterables.append(
             self.first.iter_write( 
@@ -152,7 +187,7 @@ class Pipe( object ):
     def __gt__( self, other ):
         """Pipe our output into a file
         """
-        if isinstance( other, (str,unicode)):
+        if isinstance( other, (bytes,unicode)):
             if other not in ('','-'):
                 return self.__or__( open(other,'wb') )
             else:
@@ -161,7 +196,7 @@ class Pipe( object ):
             return self.__or__( other )
     def __lt__( self, other ):
         """Pipe input from a named file"""
-        if isinstance( other, (str,unicode)):
+        if isinstance( other, (bytes,unicode)):
             if other not in ('','-'):
                 return self.__ror__( open(other,'rb') )
             else:
@@ -182,7 +217,7 @@ class Pipe( object ):
             pass 
         elif isinstance( other, file ):
             other = FileComponent( other )
-        elif isinstance( other, (str,unicode)):
+        elif isinstance( other, (bytes,unicode)):
             if other == '-':
                 other = FileComponent( sys.stdout, sys.stdin )
             elif other == '':
@@ -191,7 +226,7 @@ class Pipe( object ):
                 other = IterComponent( [other] )
         elif hasattr( other, '__iter__'):
             other = IterComponent( other )
-        elif callable( other ):
+        elif isinstance(callable, collections.Callable):
             other = FunctionComponent( other )
         if not isinstance( other, PipeComponent ):
             raise TypeError( """Require a PipeComponent-compatible object, got: %s"""%(other,) )
@@ -201,7 +236,7 @@ class Pipe( object ):
             if hasattr( process, 'kill' ):
                 try:
                     process.kill()
-                except Exception, err:
+                except Exception:
                     log.warn(
                         "Unable to kill process: %s", process,
                     )
@@ -224,9 +259,9 @@ class PipeComponent( object ):
                     elif child_result:
                         something_happened += 1
                         yield child_result
-                except StopIteration, err:
+                except StopIteration as err:
                     exhausted.append( iterable )
-                except Exception, err:
+                except Exception as err:
                     err.args += (self,)
                     raise 
             for done in exhausted:
@@ -309,7 +344,7 @@ class Process( PipeComponent ):
         named -- passed to the subprocess.Popen() command 
         
         """
-        if isinstance( command, (str,unicode)):
+        if isinstance( command, (bytes,unicode)):
             if not 'shell' in named:
                 named['shell'] = True 
         self.command = command 
@@ -319,36 +354,100 @@ class Process( PipeComponent ):
             self.good_exit = [0]
         if 'by_line' in named:
             self.by_line = named.pop( 'by_line' )
-        self.pipe = self.start_pipe(stdin,stdout,stderr, **named)
+        named['stdin'],named['stdout'],named['stderr'] = stdin,stdout,stderr
+        self.named_args = named
+        try:
+            self.no_descendent_kill = named.pop( 'no_descendent_kill' )
+        except KeyError:
+            self.no_descendent_kill = False
         super( Process, self ).__init__( )
     def __unicode__( self ):
         return u'%s( %r )'%( self.__class__.__name__, self.command )
     __repr__ = __unicode__
+    
     @property
     def stderr( self ):
         return self.pipe.stderr
+    @stderr.setter
+    def stderr( self, new ):
+        if self._pipe:
+            raise ProcessError( 'Cannot set stderr after pipe start' )
+        self.named_args['stderr'] = new 
     @property
     def stdout( self ):
         return self.pipe.stdout
+    @stdout.setter
+    def stdout( self, new ):
+        if self._pipe:
+            raise ProcessError( 'Cannot set stdout after pipe start' )
+        self.named_args['stdout'] = new 
     @property
     def stdin( self ):
         return self.pipe.stdin
+    @stdin.setter
+    def stdin( self, new ):
+        if self._pipe:
+            raise ProcessError( 'Cannot set stdin after pipe start' )
+        self.named_args['stdin'] = new 
+
+    use_blocking_stdout = False
+    def blocking_stdout( self ):
+        self.use_blocking_stdout = True
+    def direct_pipe( self, file, key='stdout' ):
+        """Directly pipe stdout/stderr/stdin to/from the file"""
+        setattr( self, key, file )
+        setattr( self, '%s_wanted'%(key,), True )
+        self.iterables.append( self.iter_wait() )
+        return file
     
-    def start_pipe( self, stdin,stdout,stderr, **named ):
+    @property 
+    def pid( self ):
+        return self.pipe.pid
+    @property 
+    def started( self ):
+        return self._pipe is not None
+    _pipe = None
+    @property 
+    def pipe( self ):
+        if self._pipe is None:
+            self._pipe = self._start_pipe( 
+                **self.named_args 
+            )
+        return self._pipe
+    def start(self):
+        """Trigger to start the process immediately"""
+        # yes, side-effects are ick...
+        self.pipe
+        return self
+    
+    def _start_pipe( self, **named ):
         """Start the captive process (internal operation)"""
-        err = None
-        if stderr == -1:
-            stderr = subprocess.STDOUT
-            stdout = True 
-        else:
-            stderr = [None,subprocess.PIPE][bool(stderr)]
+        for key in ('stdin','stdout','stderr'):
+            value = named.get(key)
+            if value == -1:
+                value = subprocess.STDOUT
+            elif hasattr( value, 'fileno' ):
+                log.info( 'Using existing file: %s', value )
+                value = value
+            else:
+                value = [None,subprocess.PIPE][bool(value)]
+            named[key] = value 
+        
         pipe = subprocess.Popen(
             self.command, 
-            stdin = [None,subprocess.PIPE][bool(stdin)],
-            stdout = [None,subprocess.PIPE][bool(stdout)],
-            stderr = [None,subprocess.PIPE][bool(stderr)],
             **named
         )
+        for key in ('stdin','stdout','stderr'):
+            if not hasattr(named[key],'fileno'):
+                if key == 'stdout' and self.use_blocking_stdout:
+                    log.info( 'Using blocking %s', key )
+                    continue 
+                    #pass
+                fh = getattr( pipe, key, None )
+                if fh is not None:
+                    set_non_blocking( fh )
+            else:
+                log.info( 'Using blocking %s', key )
         return pipe
     def __iter__( self ):
         """Iterate over the results of the process (normally done by the Pipe)"""
@@ -363,7 +462,7 @@ class Process( PipeComponent ):
             # nothing has been hooked to stderr
             if self.stderr:
                 self.iterables.append(
-                    self.iter_write( self.stderr )
+                    self.iter_read_err( )
                 )
         if not self.stdin_needed:
             # nothing is being sent in, so input is finished...
@@ -404,6 +503,9 @@ class Process( PipeComponent ):
         """Create a Pipe and run it with just this item as its children"""
         pipe = Pipe( self )
         return pipe( *args, **named )
+        
+    def poll( self ):
+        return self.pipe.poll()
     
     def check_exit( self ):
         """Check our exit code"""
@@ -411,8 +513,20 @@ class Process( PipeComponent ):
             exitcode = self.pipe.poll()
         else:
             exitcode = self.pipe.returncode
-        if exitcode < 0:
-            self.pipe.kill()
+        if exitcode is None or exitcode < 0:
+            try:
+                if not self.no_descendent_kill:
+                    from fussy.processcontrol import kill_children
+                    kill_children( self.pipe.pid )
+                else:
+                    self.kill()
+                    
+            except OSError as err:
+                if err.errno == errno.ESRCH:
+                    # we exited...
+                    pass 
+                else:
+                    raise
             exitcode = self.pipe.poll()
         if exitcode is None:
             log.error(
@@ -423,6 +537,7 @@ class Process( PipeComponent ):
                 "Process %s returned error code %s"%( self.command, exitcode, ),
             )
             err.process = self 
+            err.exitcode = exitcode
             raise err
 
     def iter_read( self ):
@@ -432,6 +547,12 @@ class Process( PipeComponent ):
         if self.by_line:
             output = by_line( output )
         return output
+    def iter_read_err( self ):
+        self.stderr_needed = True 
+        output = reader( self.stderr )
+        if self.by_line:
+            output = by_line( output )
+        return output
     def iter_write( self, source ):
         """Create a thing which will read from source and write to us"""
         self.stdin_needed = True
@@ -439,11 +560,14 @@ class Process( PipeComponent ):
             source,
             self.stdin,
         )
+    def iter_wait( self ):
+        while self.pipe.poll() == None:
+            yield DID_NOTHING
     def kill( self ):
         """Kill our underlying subprocess.Popen"""
         try:
             return self.pipe.kill( )
-        except OSError, err:
+        except OSError:
             # we've already died/closed...
             pass 
     
@@ -505,22 +629,26 @@ def writeiter( iterator, fh ):
     .. code-block:: python
     
         writeiter( 
-            response.iter_content( 4096, decode_unicode=False ),
+            response.iter_content( DEFAULT_READ_BLOCK, decode_unicode=False ),
             pipe 
         )
     """
     total_written = 0
     assert hasattr( iterator, '__iter__'), iterator
     for content in iterator:
-        if content and isinstance( content, (str,unicode)):
+        if content and isinstance( content, (bytes,unicode)):
             for block_written in writer( content, fh ):
-                total_written += block_written
-                yield None
+                if block_written is DID_NOTHING:
+                    yield block_written
+                else:
+                    total_written += block_written
+                    yield None
         else:
             yield DID_NOTHING
+    #log.debug( 'Closing %s after %s bytes', fh, total_written )
     close( fh )
 
-def writer( content,fh, encoding='utf8' ):
+def writer( content,fh, encoding='utf8', block_size=4096 ):
     """Continue writing content (string) to fh until content is consumed
     
     Used by writeiter to writing individual bits of content to the fh
@@ -528,27 +656,36 @@ def writer( content,fh, encoding='utf8' ):
     fp = fileno( fh )
     if isinstance( content, unicode ):
         content = content.encode( encoding )
-    assert isinstance( content, str ), """Can only write string values: %r"""%(content)
-    while content:
-        try:
-            written = os.write( fp, content )
-        except ValueError, err:
-            log.warn( 'Unconsumed content: %s', content )
-            break
-        if not written:
-            yield DID_NOTHING
-        else:
-            content = content[written:]
-            yield written
+    assert isinstance( content, bytes ), """Can only write byte values: %r"""%(content)
+    for block in string_blocks( content, block_size=block_size ):
+        while block:
+            written = 0
+            try:
+                written = os.write( fp, block )
+            except OSError as err:
+                if err.errno == errno.EWOULDBLOCK:
+                    yield DID_NOTHING 
+                else:
+                    log.warn( "OS Error writing: %s", err )
+                    return
+            except (IOError,ValueError) as err:
+                log.warn( "Error while writing: %s", err )
+                return
+            if not written:
+                #log.debug( 'Unable to write' )
+                yield DID_NOTHING
+            else:
+                #log.debug( 'Wrote: %s', written )
+                block = block[written:]
+                yield written
         
-def reader( fh, blocksize = 4096 ):
+def reader( fh, blocksize = DEFAULT_READ_BLOCK ):
     """Produce content blocks from fh without blocking
     """
     try:
         fd = fileno( fh )
-    except ValueError, err:
+    except ValueError as err:
         return 
-    total = 0
     while not fh.closed:
         try:
             rr,rw,re = select.select([fh],[],[],0)
@@ -558,9 +695,9 @@ def reader( fh, blocksize = 4096 ):
                     break
             else:
                 result = DID_NOTHING
-        except ValueError, err:
+        except ValueError as err:
             break 
-        except (IOError,OSError), err:
+        except (IOError,OSError) as err:
             if err.args[0] in ( errno.EWOULDBLOCK,):
                 if fh.closed:
                     break
@@ -607,3 +744,16 @@ def by_line( iterable ):
                     yield line 
     if buffer:
         yield buffer 
+
+def string_blocks( content, block_size=4096 ):
+    """Yield blocks from content only producing one block at a time"""
+    current = 0
+    length = len(content)
+    if len(content) < block_size:
+        # no point copying it...
+        yield content 
+    else:
+        while current < length:
+            new = current + block_size
+            yield content[current:new]
+            current = new
