@@ -1,7 +1,7 @@
 #! /usr/bin/env python
 """Bundle a package as a signed firmware for installation/redistribution
 """
-import tempfile, subprocess, os, sys, datetime, glob
+import tempfile, subprocess, os
 from optparse import OptionParser
 
 DEFAULT_EXCLUDES = [
@@ -78,5 +78,5 @@ def main():
     """Main function for the packing script"""
     options,args = get_options().parse_args()
     gpg_file = pack( options.root, options.exclude, options.encrypt )
-    print gpg_file 
+    print(gpg_file) 
     return 0
